/*
**   Modification of Fizick's yadif avisynth filter.
**   Copyright (C) 2007 Kevin Stone
**   Yadif C-plugin for Avisynth 2.5 - Yet Another DeInterlacing Filter
**   Copyright (C) 2007 Alexander G. Balakhnin aka Fizick  http://avisynth.org.ru
**   Port of YADIF filter from MPlayer
**   Copyright (C) 2006 Michael Niedermayer <michaelni@gmx.at>
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "yadifmod.h"

yadifmod::yadifmod(PClip _child, int _order, int _field, int _mode,
	PClip _edeint, int _opt, IScriptEnvironment *env) : GenericVideoFilter(_child), 
	order(_order), field(_field), mode(_mode), edeint(_edeint), opt(_opt)
{
	srcPF = NULL;
	dstPF = NULL;
	edeintPF = NULL;
	if (!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("yadifmod:  only yv12 and yuy2 input are supported!");
	if (order < -1 || order > 1)
		env->ThrowError("yadifmod:  order must be set to -1, 0, or 1!");
	if (field < -1 || field > 1)
		env->ThrowError("yadifmod:  field must be set to -1, 0, or 1!");
	if (mode < 0 || mode > 3)
		env->ThrowError("yadifmod:  mode must be set to 0, 1, 2, or 3!");
	if (opt < 0 || opt > 2)
		env->ThrowError("yadifmod:  opt must be set to 0, 1, or 2!");
	if (!edeint)
		env->ThrowError("yadifmod:  an edeint clip must be specified!");
	VideoInfo vi2 = edeint->GetVideoInfo();
	if (vi2.height != vi.height)
		env->ThrowError("yadifmod:  edeint clip's height doesn't match!");
	if (vi2.width != vi.width)
		env->ThrowError("yadifmod:  edeint clip's width doesn't match!");
	if (!vi.IsSameColorspace(vi2))
		env->ThrowError("yadifmod:  edeint clip's colorspace doesn't match!");
	if (vi2.num_frames != vi.num_frames*((mode&1)?2:1))
		env->ThrowError("yadifmod:  edeint clip's number of frames doesn't match!");
	edeint->SetCacheHints(CACHE_NOTHING, 0);
	child->SetCacheHints(CACHE_NOTHING, 0);
	vi_saved = vi;
	if (mode&1)
	{
		vi.num_frames *= 2;
		vi.SetFPS(vi.fps_numerator*2,vi.fps_denominator);
	}
	if (order == -1)
		order = child->GetParity(0) ? 1 : 0;
	if (field == -1)
		field = order;
	srcPF = (PFC**)calloc(3,sizeof(PFC*));
	for (int i=0; i<3; ++i)
	{
		srcPF[i] = (PFC*)calloc(1,sizeof(PFC));
		srcPF[i]->pf = new PlanarFrame(vi);
		srcPF[i]->n = -20;
	}
	dstPF = new PlanarFrame(vi);
	edeintPF = new PlanarFrame((VideoInfo)edeint->GetVideoInfo());
}

yadifmod::~yadifmod()
{
	if (dstPF)
		delete dstPF;
	if (edeintPF)
		delete edeintPF;
	if (srcPF)
	{
		for (int i=0; i<3; ++i)
		{
			if (srcPF[i])
			{
				delete srcPF[i]->pf;
				free(srcPF[i]);
			}
		}
		free(srcPF);
	}
}

void yadifmod::resetFrames(PFC **p, int start, int stop)
{
	for (int i=start; i<=stop; ++i)
	{
		for (int j=0; j<3; ++j)
		{
			if (p[j]->n == i && j != i-start)
			{
				PFC *temp = p[i-start];
				p[i-start] = p[j];
				p[j] = temp;
				break;
			}
		}
	}
}

int yadifmod::mapn(int n)
{
	if (n < 0) return 0;
	if (n >= vi_saved.num_frames) return vi_saved.num_frames-1;
	return n;
}

PVideoFrame __stdcall yadifmod::GetFrame(int n, IScriptEnvironment *env)
{
	edeintPF->copyFrom(edeint->GetFrame(n, env), (VideoInfo)edeint->GetVideoInfo());
	int fieldt = field;
	if (mode&1)
	{
		fieldt = (n&1) ? 1-order : order;
		n >>= 1;
	}
	resetFrames(srcPF,n-1,n+1);
	for (int i=n-1; i<=n+1; ++i)
	{
		if (srcPF[i-n+1]->n != i)
		{
			srcPF[i-n+1]->pf->copyFrom(child->GetFrame(mapn(i), env), vi_saved);
			srcPF[i-n+1]->n = i;
		}
	}
	const bool usemmx = opt == 0 ? ((env->GetCPUFlags()&CPUF_MMX)?true:false) :
		opt == 1 ? false : true;
	for (int b=0; b<3; ++b)
	{
		const int pitch = srcPF[0]->pf->GetPitch(b);
		const int width = srcPF[0]->pf->GetWidth(b);
		const int height = srcPF[0]->pf->GetHeight(b);
		const unsigned char *s0 = srcPF[0]->pf->GetPtr(b);
		const unsigned char *s1 = srcPF[1]->pf->GetPtr(b);
		const unsigned char *s2 = srcPF[2]->pf->GetPtr(b);
		unsigned char *dstp = dstPF->GetPtr(b);
		const unsigned char *edeintp = edeintPF->GetPtr(b);
		if (!fieldt)
			env->BitBlt(dstp,pitch,s1+pitch,pitch,width,1);
		else
			env->BitBlt(dstp+pitch,pitch,edeintp+pitch,pitch,width,1);
		env->BitBlt(dstp+(1-fieldt)*pitch,pitch*2,s1+(1-fieldt)*pitch,
			pitch*2,width,height>>1);
		const int starty = 2+fieldt;
		const int stopy = fieldt ? height-3 : height-4;
		const unsigned char *prev2pp = s0+(starty-1)*pitch;
		const unsigned char *srcpp = s1+(starty-1)*pitch;
		const unsigned char *next2pp = s2+(starty-1)*pitch;
		const unsigned char *prevp, *nextp;
		if (fieldt^order)
		{
			prevp = s1+starty*pitch;
			nextp = s2+starty*pitch;
		}
		else
		{
			prevp = s0+starty*pitch;
			nextp = s1+starty*pitch;
		}
		dstp += starty*pitch;
		edeintp += starty*pitch;
		if (usemmx)
			filter_MMX(prev2pp,prevp,srcpp,nextp,next2pp,edeintp,dstp,width,mode,pitch,
				starty, stopy);
		else
		{
			const unsigned char *prev2pn = prev2pp+pitch*2;
			const unsigned char *prevp2p = prevp-pitch*2;
			const unsigned char *prevp2n = prevp+pitch*2;
			const unsigned char *srcpn = srcpp+pitch*2;
			const unsigned char *nextp2p = nextp-pitch*2;
			const unsigned char *nextp2n = nextp+pitch*2;
			const unsigned char *next2pn = next2pp+pitch*2;
			for (int y=starty; y<=stopy; y+=2)
			{
				for (int x=0; x<width; ++x)
				{
					const int p1 = srcpp[x];
					const int p2 = (prevp[x]+nextp[x])>>1;
					const int p3 = srcpn[x];
					const int tdiff0 = abs(prevp[x]-nextp[x]);
					const int tdiff1 =(abs(prev2pp[x]-p1)+abs(prev2pn[x]-p3))>>1;
					const int tdiff2 =(abs(next2pp[x]-p1)+abs(next2pn[x]-p3))>>1;
					int diff = MAX3(tdiff0>>1, tdiff1, tdiff2);
					if (mode < 2)
					{
						const int p0 = (prevp2p[x]+nextp2p[x])>>1;
						const int p4 = (prevp2n[x]+nextp2n[x])>>1;
						const int maxs = MAX3(p2-p3, p2-p1, min(p0-p1, p4-p3));
						const int mins = MIN3(p2-p3, p2-p1, max(p0-p1, p4-p3));
						diff = MAX3(diff, mins, -maxs);
					}
					const int spatial_pred = edeintp[x];
					if (spatial_pred > p2 + diff)
						dstp[x] = p2 + diff;
					else if (spatial_pred < p2 - diff)
						dstp[x] = p2 - diff;
					else
						dstp[x] = spatial_pred;
				}
				prev2pp += pitch*2;
				prev2pn += pitch*2;
				prevp2p += pitch*2;
				prevp += pitch*2;
				prevp2n += pitch*2;
				srcpp += pitch*2;
				srcpn += pitch*2;
				nextp2p += pitch*2;
				nextp += pitch*2;
				nextp2n += pitch*2;
				next2pp += pitch*2;
				next2pn += pitch*2;
				dstp += pitch*2;
				edeintp += pitch*2;
			}
		}
		if (!fieldt)
			env->BitBlt(dstPF->GetPtr(b)+pitch*(height-2),pitch,
				edeintPF->GetPtr(b)+pitch*(height-2),pitch,width,1);
		else
			env->BitBlt(dstPF->GetPtr(b)+pitch*(height-1),pitch,s1+pitch*(height-2),
				pitch,width,1);
	}
	PVideoFrame dst = env->NewVideoFrame(vi);
	dstPF->copyTo(dst, vi);
	return dst;
}

void yadifmod::filter_MMX(const unsigned char *prev2pp, const unsigned char *prevp,
	const unsigned char *srcpp, const unsigned char *nextp, const unsigned char *next2pp,
	const unsigned char *edeintp, unsigned char *dstp, const int width, const int mode,
	const int pitch, int starty, const int stopy)
{
	const unsigned char *prev2pn = prev2pp+pitch*2;
	const unsigned char *prevp2p = prevp-pitch*2;
	const unsigned char *prevp2n = prevp+pitch*2;
	const unsigned char *srcpn = srcpp+pitch*2;
	const unsigned char *nextp2p = nextp-pitch*2;
	const unsigned char *nextp2n = nextp+pitch*2;
	const unsigned char *next2pn = next2pp+pitch*2;
	__asm
	{
		mov edx,edeintp
		mov edi,srcpp
		mov esi,srcpn
yloop:
		pxor mm7,mm7
		xor ecx,ecx
xloop:
		movd mm0,[edi+ecx]	// srcpp (p1)
		movd mm1,[esi+ecx]	// srcpn (p3)
		punpcklbw mm0,mm7
		punpcklbw mm1,mm7
		mov eax,prevp
		mov ebx,nextp
		movd mm2,[eax+ecx]	// prevp
		movd mm3,[ebx+ecx]	// nextp
		punpcklbw mm2,mm7
		punpcklbw mm3,mm7
		movq mm4,mm2		// save prevp
		paddw mm2,mm3
		psraw mm2,1			// (prevp+nextp)>>1  (p2)
		psubw mm4,mm3		// prevp-nextp
		pxor mm3,mm3
		psubw mm3,mm4
		pmaxsw mm3,mm4		// tdiff0 = abs(prevp-nextp)
		psraw mm3,1			// tdiff0>>1
		mov eax,prev2pp
		mov ebx,prev2pn
		movd mm4,[eax+ecx]
		movd mm5,[ebx+ecx]
		punpcklbw mm4,mm7
		punpcklbw mm5,mm7
		psubw mm4,mm0		// prev2pp-srcpp
		psubw mm5,mm1		// prev2pn-srcpn
		pxor mm6,mm6
		psubw mm6,mm4
		pmaxsw mm4,mm6		// abs(prev2pp-srcpp)
		pxor mm6,mm6
		psubw mm6,mm5
		pmaxsw mm5,mm6		// abs(prev2pn-srcpn)
		paddw mm4,mm5		// abs(prev2pp-p1)+abs(prev2pn-p3)
		psraw mm4,1			// tdiff1 = (abs(prev2pp-p1)+abs(prev2pn-p3))>>1
		mov eax,next2pp
		mov ebx,next2pn
		movd mm5,[eax+ecx]
		movd mm6,[ebx+ecx]
		punpcklbw mm5,mm7
		punpcklbw mm6,mm7
		psubw mm5,mm0		// next2pp-srcpp
		psubw mm6,mm1		// next2pn-srcpn
		psubw mm7,mm5
		pmaxsw mm5,mm7		// abs(next2pp-srcpp)
		pxor mm7,mm7
		psubw mm7,mm6
		pmaxsw mm6,mm7		// abs(next2pn-srcpn)
		paddw mm5,mm6		// abs(next2pp-p1)+abs(next2pn-p3)
		psraw mm5,1			// tdiff2 = (abs(next2pp-p1)+abs(next2pn-p3))>>1
		pmaxsw mm3,mm4
		pmaxsw mm3,mm5		// diff = MAX3(tdiff0>>1, tdiff1, tdiff2);
		cmp mode,2
		jge skip_spatial_check
		pxor mm7,mm7
		mov eax,prevp2p
		mov ebx,nextp2p
		movd mm4,[eax+ecx]	// prevp2p
		movd mm5,[ebx+ecx]	// nextp2p
		punpcklbw mm4,mm7
		punpcklbw mm5,mm7
		paddw mm4,mm5
		psraw mm4,1			// p0 = (prevp2p+nextp2p)>>1
		mov eax,prevp2n
		mov ebx,nextp2n
		movd mm5,[eax+ecx]	// prevp2n
		movd mm6,[ebx+ecx]	// nextp2n
		punpcklbw mm5,mm7
		punpcklbw mm6,mm7
		paddw mm5,mm6
		psraw mm5,1			// p4 = (prevp2n+nextp2n)>>1
		movq mm6,mm2
		movq mm7,mm2
		psubw mm6,mm1		// p2-p3
		psubw mm7,mm0		// p2-p1
		psubw mm5,mm1		// p4-p3
		psubw mm4,mm0		// p0-p1
		movq mm0,mm4
		movq mm1,mm4
		pminsw mm0,mm5
		pmaxsw mm1,mm5
		pmaxsw mm0,mm7
		pminsw mm1,mm7
		pmaxsw mm0,mm6		// maxs = MAX3(p2-p3, p2-p1, min(p0-p1, p4-p3))
		pminsw mm1,mm6		// mins = MIN3(p2-p3, p2-p1, max(p0-p1, p4-p3))
		pxor mm7,mm7
		pmaxsw mm3,mm1
		psubw mm7,mm0
		pmaxsw mm3,mm7		// diff = MAX3(diff, mins, -maxs)
skip_spatial_check:
		movq mm0,mm2
		paddw mm2,mm3		// p2 + diff
		psubw mm0,mm3		// p2 - diff
		pxor mm7,mm7
		movd mm1,[edx+ecx]	// edeintp
		punpcklbw mm1,mm7
		pminsw mm1,mm2
		pmaxsw mm1,mm0
		packuswb mm1,mm1
		mov eax,dstp
		movd [eax+ecx],mm1
		add ecx,4
		cmp ecx,width
		jl xloop
		mov eax,pitch
		add eax,pitch
		add prev2pp,eax
		add prev2pn,eax
		add prevp2p,eax
		add prevp,eax
		add prevp2n,eax
		add edi,eax
		add esi,eax
		add nextp2p,eax
		add nextp,eax
		add nextp2n,eax
		add next2pp,eax
		add next2pn,eax
		add dstp,eax
		add edx,eax
		mov ebx,stopy
		add starty,2
		cmp starty,ebx
		jle yloop
		emms
	}
}

AVSValue __cdecl Create_yadifmod(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	if (!args[0].IsClip() || !args[4].IsClip())
		env->ThrowError("yadifmod:  args 0 and 4 must be clips!");
	return new yadifmod(args[0].AsClip(),args[1].AsInt(-1),args[2].AsInt(-1),
		args[3].AsInt(0),args[4].AsClip(),args[5].AsInt(0),env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) 
{
	env->AddFunction("yadifmod","c[order]i[field]i[mode]i[edeint]c[opt]i", 
		Create_yadifmod, 0);
    return 0;
}